#!/usr/bin/env python
# coding: utf-8

# In[37]:


# import tkinter as tk
from tkinter import colorchooser, ttk, simpledialog
from tkinter.filedialog import askopenfilename, asksaveasfilename
from tkinter import messagebox
from PIL import Image, ImageTk
import math
import os

# Layout constants.
MARGIN = 5
WINDOW_WID = 400
WINDOW_HGT = 300

class ScribbleDocumentsApp:
    # Create and manage the tkinter interface.
    def __init__(self):
        # Initially there is no document name and it hasn't been modified.
        self.document_name = ''
        self.document_title = ''
        self.is_modified = False

        # Load the MRU list.
        self.max_mru_length = 4
        self.load_mru_list()

        # Make the main interface.
        self.window = tk.Tk()
        self.window.title('scribble_documents')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry(f'{WINDOW_WID}x{WINDOW_HGT}')

        # Build menus.
        self.build_menus()

        # Build the toolbar.
        self.canvas = None
        self.build_toolbar()

        # Build the rest of the UI here...
        self.canvas = tk.Canvas(self.window, bg='white',
            borderwidth=2, relief=tk.SUNKEN)
        self.canvas.pack(padx=10, pady=(0, 10),
            side=tk.BOTTOM, fill=tk.BOTH, expand=True)

        # Initially we are not drawing anything.
        self.line_points = None
        self.current_line = None

        self.start_point = None
        self.end_point = None
        self.current_oval = None
        self.current_rectangle = None
        self.current_polygon = None
        self.current_star = None
        self.current_text = None

        # Initially select the arrow.
        self.select_tool('tool_arrow')

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def build_menus(self):
        menu_bar = tk.Menu(self.window)
        self.window.config(menu=menu_bar)

        # File menu
        self.build_file_menu(menu_bar)

        # MRU menu
        self.build_mru_menu()

    def build_file_menu(self, menu_bar):
        '''Build the File menu.'''
        file_menu = tk.Menu(menu_bar, tearoff=False)
        self.file_menu = file_menu    # Save for use by the MRU list code.
        menu_bar.add_cascade(label='File', menu=file_menu)

        # Note: You must save the photo image in a persistent
        # variable so tkinter can redraw it when needed.

        # Note: You need to include the compound parameter
        # if you want to use both images and text. That value
        # tells where to put the image relative to the text.

        # New
        self.new_photo_image = ImageTk.PhotoImage(file='new.png')        
        file_menu.add_command(label='New', accelerator='Ctrl+N', compound='left',
            command=self.mnu_new, image=self.new_photo_image)
        self.window.bind('<Control-n>', lambda event: self.mnu_new())

        # Open
        self.open_photo_image = ImageTk.PhotoImage(file='open.png')
        file_menu.add_command(label='Open...', accelerator='Ctrl+O', compound='left',
            command=self.mnu_open, image=self.open_photo_image)
        self.window.bind('<Control-o>', lambda event: self.mnu_open())

        file_menu.add_separator()

        # Note: If you don't include an image, the menu item doesn't
        # quite line up with items that do have an image. You can fix
        # it by including a blank image.

        # Save
        self.save_photo_image = ImageTk.PhotoImage(file='save.png')
        file_menu.add_command(label='Save', accelerator='Ctrl+S', compound='left',
            command=self.mnu_save, image=self.save_photo_image)
        self.window.bind('<Control-s>', lambda event: self.mnu_save())

        # Save As
        self.save_as_photo_image = ImageTk.PhotoImage(file='save_as.png')
        file_menu.add_command(label='Save As...', compound='left',
            command=self.mnu_save_as, image=self.save_as_photo_image)

        file_menu.add_separator()

        # Exit
        self.exit_photo_image = ImageTk.PhotoImage(file='exit.png')
        file_menu.add_command(label='Exit', compound='left',
            command=self.mnu_exit, image=self.exit_photo_image)

    def mnu_new(self):
        # Make sure it's safe to clear the document.
        if not self.is_safe_to_close(): return
        self.canvas.delete(tk.ALL)
        self.document_name = ''
        self.document_title = ''
        self.set_is_modified(False)

    def mnu_open(self):
        # Make sure it's safe to clear the document.
        if not self.is_safe_to_close(): return

        filetypes = [
            ('JSON Files', '*.json'),
            ('All Files', '*,*')]
        filename = askopenfilename(filetypes=filetypes, defaultextension='.json')
        if len(filename) == 0: return

        self.open_file(filename)

    def mru_file_selected(self, filename):
        self.open_file(filename)

    def open_file(self, filename):
        try:
            with open(filename, 'r') as file:
                json_data = file.read()
            deserialize_canvas(self.canvas, json_data)
        except Exception as e:
            messagebox.showerror('Error',
                f'Error opening file {filename},\n{e}')

            # Remove from the MRU list.
            self.mru_remove_file(filename)
            return

        # Update the document name.
        self.document_name = filename
        self.document_title = os.path.basename(filename)
        self.set_is_modified(False)

        # Move to the top of the MRU list.
        self.mru_add_file(self.document_name)        

    def mnu_save(self):
        '''Save the document with the current file name.'''   
        if self.document_name == '':
            # We have no file name. Ask the user to pick one.
            self.mnu_save_as()
        else:
            # Save with the current file name.
            self.save_document(self.document_name)

    def mnu_save_as(self):
        filetypes = [
            ('JSON Files', '*.json'),
            ('All Files', '*,*')]
        filename = asksaveasfilename(filetypes=filetypes, defaultextension='.json')
        if len(filename) == 0: return
        self.save_document(filename)

    def save_document(self, filename):
        '''Save the current document in {filename}.'''
        try:
            # Save the serialization.
            json_data = serialize_canvas(self.canvas)
            with open(filename, 'w', encoding='utf-8') as f:
                f.write(json_data)

            # Update the document name.
            self.document_name = filename
            self.document_title = os.path.basename(filename)
            self.set_is_modified(False)

            # Move to the top of the MRU list.
            self.mru_add_file(self.document_name)

        except Exception as e:
            messagebox.showerror('Error',
                f'Error saving file {filename},\n{e}')

            # Remove from the MRU list.
            self.mru_remove_file(self.document_name)

    def mnu_exit(self):
        self.kill_callback()

    def load_mru_list(self):
        '''Load the MRU list.'''
        self.mru_list = []
        try:
            # Load the MRU list JSON file.
            with open('mru_list.json', 'r') as f:
                file_dict = json.load(f)

            # Pull the file names out of the loaded data.
            for i in range(self.max_mru_length):
                key = f'MRU_{i}'
                if key in file_dict:
                    self.mru_list.append(file_dict[key])
        except:
            pass

    def build_mru_menu(self):
        # MRU menu
        self.mru_menu = tk.Menu(self.file_menu, tearoff=False)
        self.mru_list_photo_image = ImageTk.PhotoImage(file='mru_list.png')
        self.file_menu.add_cascade(label='Recent Files', menu=self.mru_menu,
            compound='left', image=self.mru_list_photo_image)

        # Make the initial MRU list.
        self.document_photo_image = ImageTk.PhotoImage(file='document.png')
        self.build_mru_list()

    def build_mru_list(self):
        '''Add files to the MRU menu.'''
        # Delete existing items.
        self.mru_menu.delete(0, tk.END)

        # Add the files.
        # Note how the lambda function renames the variable.
        # This forces it to bind the value when this code is
        # executed rather than trying to use the value of
        # filename when the lambda executes.
        for filename in self.mru_list:
            label = os.path.basename(filename)
            self.mru_menu.add_command(label=label, compound='left',
                command=lambda bound_name=filename: self.mru_file_selected(bound_name),
                image=self.document_photo_image)

        # Enable the Recent Files command if there are no recent files.
        if len(self.mru_list) > 0:
            state = 'normal'
        else:
            state = 'disabled'
        self.file_menu.entryconfigure('Recent Files', state=state)

        # Save the MRU list.
        self.save_mru_list()

    def save_mru_list(self):
        '''Save the MRU list.'''
        # Build a dictionary holding the MRU file names.
        file_dict = {}
        for i in range(len(self.mru_list)):
            key = f'MRU_{i}'
            file_dict[key] = self.mru_list[i]

        # Save the JSON file.
        with open('mru_list.json', 'w') as f:
            json.dump(file_dict, f)

    def mru_add_file(self, filename):
        '''Add the file name to the MRU list.'''
        # Move the file name to the top of the list.
        if filename in self.mru_list:
            self.mru_list.remove(filename)
        self.mru_list.insert(0, filename)

        # limit the number of names on the list.
        self.mru_list = self.mru_list[:self.max_mru_length]

        # Rebuild the MRU menu.
        self.build_mru_list()

    def mru_remove_file(self, filename):
        '''Remove the file name from the MRU list.'''
        if filename in self.mru_list:
            self.mru_list.remove(filename)

        # Rebuild the MRU menu.
        self.build_mru_list()

    def build_toolbar(self):
        '''Build the toolbar.'''
        self.tool_hgt = 16
        self.tool_wid = 16

        # Create a toolbar
        self.toolbar = tk.Frame(self.window, relief=tk.RIDGE, borderwidth=2)

        # Add buttons and other tools to the toolbar.
        self.tool_padx = self.tool_pady = 3

        # Outline color button.
        self.make_outline_button()

        # Fill color button.
        self.make_fill_button()

        # Line thickness OptionMenu.
        self.make_thickness_option_menu()

        # Line dash OptionMenu.
        self.make_dash_option_menu()

        # Tools.
        self.make_tools()

        # Place the toolbar
        self.toolbar.pack(side=tk.TOP, fill=tk.X, padx=5, pady=5)

    def make_outline_button(self):
        self.outline_button_image = tk.PhotoImage(width=self.tool_wid, height=self.tool_hgt)
        self.outline_button = tk.Button(self.toolbar, command=self.choose_outline_color, image=self.outline_button_image)
        self.outline_button.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)
        self.set_outline_color('black')

    def make_fill_button(self):
        self.fill_button_image = tk.PhotoImage(width=self.tool_wid, height=self.tool_hgt)
        self.fill_button = tk.Button(self.toolbar, command=self.choose_fill_color, image=self.fill_button_image)
        self.fill_button.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)
        self.set_fill_color('white')

    def make_thickness_option_menu(self):
        # Make a dictionary mapping thickness values to images.
        # Note: When I tried using strings holding the numbers 1 through 5
        # or integers holding the numbers 1 through 5, the menu seemed to get
        # confused so the first menu item showed the text '1'.
        # Hence you need to add 1 to the selected option to get the
        # desired line thickness.
        self.thickness_images = {
            0: tk.PhotoImage(file='thickness_1.png'),
            1: tk.PhotoImage(file='thickness_2.png'),
            2: tk.PhotoImage(file='thickness_3.png'),
            3: tk.PhotoImage(file='thickness_4.png'),
            4: tk.PhotoImage(file='thickness_5.png'),
        }

        # Make the OptionMenu.
        self.width_var = tk.IntVar(value=0)
        self.thickness_om = tk.OptionMenu(self.toolbar,
            self.width_var, 0, 1, 2, 3, 4,
            command=self.thickness_selected)
        self.thickness_om.config(indicatoron=False)
        thickness = self.width_var.get()
        self.thickness_om.configure(image=self.thickness_images[thickness])

        # Get the underlying menu.
        menu = self.thickness_om['menu']

        # Add the menu item images.
        for number, image in self.thickness_images.items():
            menu.entryconfigure(number, image=image, compound=tk.NONE)

        self.thickness_om.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)

    def make_dash_option_menu(self):
        # Not all dash patterns produce results on all platforms. See:
        # https://stackoverflow.com/questions/41796792/tkinter-canvas-dash-option-is-not-behaving-as-expected
        # Make a dictionary mapping dash values to images.
        self.dash_images = {
            'dash_solid':       tk.PhotoImage(file='dash_solid.png'),
            'dash_6_4':         tk.PhotoImage(file='dash_6_4.png'),
            'dash_6_4_2_4':     tk.PhotoImage(file='dash_6_4_2_4.png'),
            'dash_6_4_2_4_2_4': tk.PhotoImage(file='dash_6_4_2_4_2_4.png'),
        }

        # Make the OptionMenu.
        self.dash_var = tk.StringVar(value='dash_solid')
        self.dash_om = tk.OptionMenu(self.toolbar, self.dash_var,
            'dash_solid', 'dash_6_4', 'dash_6_4_2_4', 'dash_6_4_2_4_2_4',
            command=self.dash_selected)
        self.dash_om.config(indicatoron=False)
        self.dash_om.configure(image=self.dash_images['dash_solid'])

        # Get the underlying menu.
        menu = self.dash_om['menu']

        # Add the menu item images.
        for name, image in self.dash_images.items():
            menu.entryconfigure(name, image=image, compound=tk.NONE)

        self.dash_om.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)

        # Start with solid lines.
        self.dash_pattern = ()

    def make_tools(self):
        separator = ttk.Separator(self.toolbar, orient=tk.VERTICAL)
        separator.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady, fill=tk.Y)

        self.tool_images = {
            'tool_arrow':     tk.PhotoImage(file='tool_arrow.png'),
            'tool_scribble':  tk.PhotoImage(file='tool_scribble.png'),
            'tool_oval':      tk.PhotoImage(file='tool_oval.png'),
            'tool_rectangle': tk.PhotoImage(file='tool_rectangle.png'),
            'tool_polygon':   tk.PhotoImage(file='tool_polygon.png'),
            'tool_star':      tk.PhotoImage(file='tool_star.png'),
            'tool_text':      tk.PhotoImage(file='tool_text.png'),
        }
        self.tool_buttons = {}
        for name, image in self.tool_images.items():
            button = tk.Button(self.toolbar, image=image,
                command=lambda bound_name=name: self.select_tool(bound_name))
            button.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)
            self.tool_buttons[name] = button

    def kill_callback(self):
        # Make sure it's safe to close the program.
        if not self.is_safe_to_close(): return
        self.window.destroy()

    def choose_outline_color(self):
        color = self.outline_color
        if color == '': color = 'white' 

        color = colorchooser.askcolor(title='Outline Color',
            initialcolor=color)
        if color[1] is None:
            self.set_outline_color('')
        else:
            self.set_outline_color(color[1])

    def set_outline_color(self, color):
        # Save the selected color.
        self.outline_color = color

        # Draw the selected color on the button.
        y1 = int(self.tool_hgt / 2)
        y2 = y1 + 2
        bg_color = '#F0F0F0'
        self.outline_button_image.put(bg_color, to=(0,y1,self.tool_wid-1,y2))
        self.outline_button_image.put(color, to=(0,y1,self.tool_wid-1,y2))

    def choose_fill_color(self):
        color = self.fill_color
        if color == '': color = 'white'

        color = colorchooser.askcolor(title='Fill Color',
            initialcolor=color)
        if color[1] is None:
            self.set_fill_color('')
        else:
            self.set_fill_color(color[1])

    def set_fill_color(self, color):
        # Save the selected color.
        self.fill_color = color

        # Draw the selected color on the button.
        x1 = 1
        x2 = self.tool_wid - 1
        y1 = 3
        y2 = self.tool_hgt - 4
        self.fill_button_image.put('black', to=(x1,y1,x2,y2))
        if color == '': color = '#F0F0F0'
        self.fill_button_image.put(color, to=(x1+1,y1+1,x2-1,y2-1))

    def thickness_selected(self, option_num):
        # Give the OptionMenu the correct image.
        self.thickness_om.configure(image=self.thickness_images[option_num])

    def dash_selected(self, option_name):
        # Save the selected dash pattern.
        match self.dash_var.get():
            case 'dash_solid':
                self.dash_pattern = ()
            case 'dash_6_4':
                self.dash_pattern = (6, 4)
            case 'dash_6_4_2_4':
                self.dash_pattern = (6, 4, 2, 4)
            case 'dash_6_4_2_4_2_4':
                self.dash_pattern = (6, 4, 2, 4, 2, 4)

        # Give the OptionMenu the correct image.
        self.dash_om.configure(image=self.dash_images[option_name])

    def scribble_mouse_down(self, event):
        # Save the starting point.
        self.line_points = [(event.x, event.y)]

    def scribble_mouse_move(self, event):
        # Ignore extraneous events.
        if self.line_points is None: return

        # If there is a current line, remove it.
        if self.current_line is not None:
            self.canvas.delete(self.current_line)

        # Create a line.
        # Note: We use the outline color for the line's fill parameter.
        self.line_points.append((event.x, event.y))
        width, dash, outline, fill = self.get_drawing_parameters()
        self.current_line = self.canvas.create_line(self.line_points,
            width=width, dash=dash, fill=outline)

        # Mark the document as modified.
        self.set_is_modified(True)

    def scribble_mouse_up(self, event):
        # Ignore extraneous events.
        if self.line_points is None: return

        # We are no longer creating a line.
        self.line_points = None
        self.current_line = None

    def oval_mouse_down(self, event):
        # Save the starting point.
        self.start_point = (event.x, event.y)

    def oval_mouse_move(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # If there is a current oval, remove it.
        if self.current_oval is not None:
            self.canvas.delete(self.current_oval)

        # Create an oval.
        self.end_point = (event.x, event.y)
        width, dash, outline, fill = self.get_drawing_parameters()
        self.current_oval = self.canvas.create_oval(
            self.start_point, self.end_point,
            width=width, dash=dash, outline=outline, fill=fill)

        # Mark the document as modified.
        self.set_is_modified(True)

    def oval_mouse_up(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # We are no longer creating an oval.
        self.start_point = None
        self.end_point = None
        self.current_oval = None

    def rectangle_mouse_down(self, event):
        # Save the starting point.
        self.start_point = (event.x, event.y)

    def rectangle_mouse_move(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # If there is a current rectangle, remove it.
        if self.current_rectangle is not None:
            self.canvas.delete(self.current_rectangle)

        # Create an rectangle.
        self.end_point = (event.x, event.y)
        width, dash, outline, fill = self.get_drawing_parameters()
        self.current_rectangle = self.canvas.create_rectangle(
            self.start_point, self.end_point,
            width=width, dash=dash, outline=outline, fill=fill)

        # Mark the document as modified.
        self.set_is_modified(True)

    def rectangle_mouse_up(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # We are no longer creating an rectangle.
        self.start_point = None
        self.end_point = None
        self.current_rectangle = None

    def polygon_mouse_down(self, event):
        # Save the starting point.
        self.start_point = (event.x, event.y)

    def polygon_mouse_move(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # If there is a current polygon, remove it.
        if self.current_polygon is not None:
            self.canvas.delete(self.current_polygon)

        # Create a polygon.
        self.end_point = (event.x, event.y)
        width, dash, outline, fill = self.get_drawing_parameters()
        self.draw_polygon(self.num_polygon_sides,
            self.start_point, self.end_point,
            width=width, dash=dash, outline=outline, fill=fill)

        # Mark the document as modified.
        self.set_is_modified(True)

    def draw_polygon(self, num_polygon_sides, start_point, end_point, width, dash, outline, fill):
        cx = (start_point[0] + end_point[0]) / 2
        cy = (start_point[1] + end_point[1]) / 2
        rx = (start_point[0] - end_point[0]) / 2
        ry = (start_point[1] - end_point[1]) / 2
        points = []
        theta = math.pi / 2    # Put the first vertex at the top.
        dtheta = 2 * math.pi / self.num_polygon_sides
        theta_limit = theta + 2 * math.pi - dtheta / 2
        for i in range(self.num_polygon_sides):
            points.append((
                cx + rx * math.cos(theta),
                cy + ry * math.sin(theta),
            ))
            theta += dtheta
        self.current_polygon = self.canvas.create_polygon(points,
            width=width, dash=dash, outline=outline, fill=fill)

    def polygon_mouse_up(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # We are no longer creating an polygon.
        self.start_point = None
        self.end_point = None
        self.current_polygon = None

    def star_mouse_down(self, event):
        # Save the starting point.
        self.start_point = (event.x, event.y)

    def star_mouse_move(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # If there is a current star, remove it.
        if self.current_star is not None:
            self.canvas.delete(self.current_star)

        # Create a star.
        self.end_point = (event.x, event.y)
        width, dash, outline, fill = self.get_drawing_parameters()
        self.draw_star(self.num_star_points,
            self.start_point, self.end_point,
            width=width, dash=dash, outline=outline, fill=fill)

        # Mark the document as modified.
        self.set_is_modified(True)

    def draw_star(self, num_star_points, start_point, end_point, width, dash, outline, fill):
        cx = (start_point[0] + end_point[0]) / 2
        cy = (start_point[1] + end_point[1]) / 2
        rx1 = (start_point[0] - end_point[0]) / 2
        ry1 = (start_point[1] - end_point[1]) / 2
        rx2 = rx1 * 0.5
        ry2 = ry1 * 0.5
        points = []
        theta = math.pi / 2    # Put the first vertex at the top.
        dtheta = 2 * math.pi / self.num_star_points / 2
        for i in range(self.num_star_points):
            points.append((
                cx + rx1 * math.cos(theta),
                cy + ry1 * math.sin(theta),
            ))
            theta += dtheta

            points.append((
                cx + rx2 * math.cos(theta),
                cy + ry2 * math.sin(theta),
            ))
            theta += dtheta
        self.current_star = self.canvas.create_polygon(points,
            width=width, dash=dash, outline=outline, fill=fill)

    def star_mouse_up(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # We are no longer creating an star.
        self.start_point = None
        self.end_point = None
        self.current_star = None

    def text_mouse_down(self, event):
        # Save the starting point.
        self.start_point = (event.x, event.y)

        self.start_point = (event.x, event.y)
        self.draw_text(self.start_point, outline=self.outline_color)

        # Mark the document as modified.
        self.set_is_modified(True)

    def text_mouse_move(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # If there is a current text, remove it.
        if self.current_text is not None:
            self.canvas.delete(self.current_text)

        # Create a text.
        self.start_point = (event.x, event.y)
        self.draw_text(self.start_point, outline=self.outline_color)

    def draw_text(self, start_point, outline):
        self.current_text = self.canvas.create_text(
            start_point[0], start_point[1],
            text=self.current_text_string,
            fill=outline, font=('Helvetica 15 bold'))

    def text_mouse_up(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # We are no longer creating an text.
        self.start_point = None
        self.end_point = None
        self.current_text = None
        self.current_string = None

    def get_drawing_parameters(self):
        width = self.width_var.get() + 1
        return width, self.dash_pattern, self.outline_color, self.fill_color

    def select_tool(self, tool_name):
        # Save the selected tool.
        self.selected_tool = tool_name

        # Unbind the mouse events.
        self.canvas.unbind('<Button-1>')
        self.canvas.unbind('<B1-Motion>')
        self.canvas.unbind('<ButtonRelease-1>')

        # Push the selected button down.
        for name, button in self.tool_buttons.items():
            if name == tool_name:
                button.config(relief=tk.SUNKEN)
            else:
                button.config(relief=tk.RAISED)

        # Set the appropriate cursor.
        match tool_name:
            case 'tool_arrow':
                self.canvas.config(cursor='')
            case 'tool_scribble':
                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.scribble_mouse_down)
                self.canvas.bind('<B1-Motion>', self.scribble_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.scribble_mouse_up)
            case 'tool_oval':
                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.oval_mouse_down)
                self.canvas.bind('<B1-Motion>', self.oval_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.oval_mouse_up)
            case 'tool_rectangle':
                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.rectangle_mouse_down)
                self.canvas.bind('<B1-Motion>', self.rectangle_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.rectangle_mouse_up)
            case 'tool_polygon':
                # Get the number of sides the polygon should have.
                self.num_polygon_sides = simpledialog.askinteger(
                    'Polygon Sides', '# Sides:\t\t\t\t\t')
                if self.num_polygon_sides is None:
                    # Cancel.
                    self.select_tool('tool_arrow')
                    return

                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.polygon_mouse_down)
                self.canvas.bind('<B1-Motion>', self.polygon_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.polygon_mouse_up)
            case 'tool_star':
                # Get the number of points the star should have.
                self.num_star_points = simpledialog.askinteger(
                    'Star Points', '# Points:\t\t\t\t\t')
                if self.num_star_points is None:
                    # Cancel.
                    self.select_tool('tool_arrow')
                    return

                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.star_mouse_down)
                self.canvas.bind('<B1-Motion>', self.star_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.star_mouse_up)
            case 'tool_text':
                # Get the text we should draw.
                self.current_text_string = simpledialog.askstring(
                    'Text', 'Text:\t\t\t\t\t')
                if self.current_text_string is None:
                    # Cancel.
                    self.select_tool('tool_arrow')
                    return

                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.text_mouse_down)
                self.canvas.bind('<B1-Motion>', self.text_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.text_mouse_up)

    def set_is_modified(self, value):
        '''Mark the current document as modified or not.'''
        self.is_modified = value
        if self.is_modified:
            self.window.title(f'scribble_documents * [{self.document_title}]')
        else:
            self.window.title(f'scribble_documents [{self.document_title}]')

    def is_safe_to_close(self):
        '''Return True if it is safe to close the document.'''
        # If the current document has no changes, it's safe.
        if not self.is_modified: return True

        # See if we have a current document name.
        if self.document_name == '':
            prompt = 'Save changes?'
        else:
            prompt = f'Save changes to {self.document_title}?'

        # Ask if the user wants to save changes.
        result = messagebox.askyesnocancel('Unsaved Changes?', prompt)

        # If the user clicked No, discard the changes. Safe to close.
        if result is False: return True

        # If the user clicked Cancel, cancel the save. Not safe to close.
        if result is None: return False

        # Otherwise try to save the changes.
        self.mnu_save()

        # If we saved, then it's not safe to close.
        return not self.is_modified


# In[38]:


import tkinter as tk
import json

def serialize_canvas(canvas):
    items = []
    for item in canvas.find_all():
        item_data = {
            'type':    canvas.type(item),
            'coords':  canvas.coords(item),
            'options': {}
        }
        for option in canvas.itemconfig(item):
            value = canvas.itemcget(item, option)
            if value != '':
                item_data['options'][option] = value
        items.append(item_data)

    # Return the serialized data.
    return json.dumps(items, indent=4)

def deserialize_canvas(canvas, data):
    # Delete any existing items.
    canvas.delete(tk.ALL)

    # Deserialize the JSON data.
    items = json.loads(data)

    # Recreate the items.
    for item in items:
        coords = item['coords']
        options = item['options']
        if item['type'] == 'line':
            canvas.create_line(*coords, **options)
        elif item['type'] == 'oval':
            canvas.create_oval(*coords, **options)
        elif item['type'] == 'rectangle':
            canvas.create_rectangle(*coords, **options)
        elif item['type'] == 'polygon':   # Includes stars.
            canvas.create_polygon(*coords, **options)
        elif item['type'] == 'text':
            canvas.create_text(*coords, **options)


# In[39]:


ScribbleDocumentsApp()
print('Done')


# In[ ]:




